@@ -9,11 +9,11 @@ import (
 	"time"
 
 	lru "github.com/hashicorp/golang-lru/v2"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage/storagemgr/partition/snapshot"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/grpc/protoregistry"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/log"
-	"gitlab.com/gitlab-org/gitaly/v16/middleware"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/storage"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/storage/storagemgr/partition/snapshot"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/grpc/protoregistry"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/log"
+	"gitlab.com/gitlab-org/gitaly/v18/middleware"
 	"google.golang.org/grpc"
 	"google.golang.org/protobuf/proto"
 )
@@ -179,7 +179,8 @@ func NewDryRunStreamInterceptor(logger log.Logger, registry *protoregistry.Regis
 func collectDryRunStatsForRPC(ctx context.Context, logger log.Logger, registry *protoregistry.Registry, locator storage.Locator, fullMethod string, req proto.Message, cache *DryRunLogCache) (returnErr error) {
 	methodInfo, err := registry.LookupMethod(fullMethod)
 	if err != nil {
-		return fmt.Errorf("lookup method: %w", err)
+		// Health check endpoints are not part of gitaly proto, we should simply ignore them.
+		return nil
 	}
 
 	// Only collect stats for repository-scoped RPCs
